/* Remove all the border and bg styling */
.thingy-button.flat,
.thingy-button.flat:hover,
.thingy-button.flat:active {
	border: none;
	background-image: none;
	background-color: transparent;
}

.thingy-image {
	border-radius: 0px;
	border: 1px solid #aaa;
	box-shadow: 4px 4px 6px alpha(black, 0.2);
	margin-bottom: 7px;
	-gtk-icon-shadow: none;
}

.thingy-button:hover .thingy-image {
	box-shadow: 4px 4px 6px alpha(black, 0.4);
}